from abc import ABCMeta, abstractmethod

import yaml
import io
import argparse
import json

class BaseLunaUtils(object):
    
    __metaclass__ = ABCMeta

    def ParseArguments(self):
        return self._args, self._userInput

    def Init(self, luna_config, run_mode, args, userInput):
        self._luna_config = luna_config
        self._run_mode = run_mode
        self._args= args
        self._userInput = userInput
        self._logger = None

    @property
    def luna_config(self):
        return self._luna_config

    @property
    def run_mode(self):
        return self.run_mode

    @property
    def args(self):
        return self._args

    @property
    def user_input(self):
        return self._userInput

    @property
    def logger(self):
        return self._logger

    @abstractmethod
    def RegisterModel(self, model_path, description, luna_python_model=None):
        """
        register model
        """

    @abstractmethod
    def DeployModel(self):
        """
        deploy model
        """

    @abstractmethod
    def DownloadModel(self, model_path):
        """
        download model
        """


from abc import ABCMeta, abstractmethod


class BaseLunaLogger(object):
    
    __metaclass__ = ABCMeta

    @abstractmethod
    def log_metric(self, key, value):
        """
        log numeric metric
        """

    @abstractmethod
    def log_metrics(self, metrics):
        """
        log numeric metrics
        """

    @abstractmethod
    def upload_artifacts(self, local_path, upload_path):
        """
        upload artifacts (files) to the run
        """


from abc import ABCMeta, abstractmethod
from luna.azuremlLunaUtils import AzureMLLunaUtils
from luna.mlflowLunaUtils import MLflowLunaUtils
from luna.localLunaUtils import LocalLunaUtils

import yaml
import io
import argparse
import json
import os

MLFLOW_MODEL_PATH = "MLFLOW_MODELS"

class LunaUtils(object):
    
    @staticmethod
    def Create(luna_config_file='luna_config.yml', run_mode='default', parameters={}):
        with open(luna_config_file) as file:
            luna_config = yaml.full_load(file)
        
        args, userInput = LunaUtils.ParseArgument()
        
        if run_mode == 'default':
            run_mode = luna_config['run_mode']

        if run_mode == 'azureml':
            utils = AzureMLLunaUtils()
        elif run_mode == 'mlflow':
            utils = MLflowLunaUtils()
        elif run_mode == 'local':
            utils = LocalLunaUtils()

        utils.Init(luna_config, run_mode, args, userInput)
        return utils

    @staticmethod
    def ParseArgument():
        
        parser = argparse.ArgumentParser("lunaOperation")

        parser.add_argument("--runMode", type=str, help="run mode")
        parser.add_argument("--userInput", type=str, help="input data")
        parser.add_argument("--userId", type=str, help="user id")
        parser.add_argument("--operationId", type=str, help="predecessor id")
        parser.add_argument("--predecessorOperationId", type=str, help="predecessor operation id")
        parser.add_argument("--productName", type=str, help="product name")
        parser.add_argument("--deploymentName", type=str, help="deployment name")
        parser.add_argument("--apiVersion", type=str, help="api version")
        parser.add_argument("--subscriptionId", type=str, help="subscription id")

        args = parser.parse_args()
        userInput = json.loads(args.userInput)
        return args, userInput

    @staticmethod
    def GetModelPath(run_mode, context):
        if run_mode == 'azureml':
            return os.getenv('AZUREML_MODEL_DIR')
        elif run_mode == 'mlflow':
            return context[MLFLOW_MODEL_PATH]
        else:
            return 'models'


from luna.logging.baseLunaLogger import BaseLunaLogger

class LocalLunaLogger(BaseLunaLogger):
    
    def log_metric(self, key, value):
        """
        log numeric metric
        """

    def log_metrics(self, metrics):
        """
        log numeric metrics
        """

    def upload_artifacts(self, local_path, upload_path):
        """
        upload artifacts (files) to the run
        """


from azureml.core import Workspace, Run, Experiment
from azureml.core.compute import ComputeTarget, AmlCompute
from azureml.core.environment import Environment
from azureml.core.model import Model, InferenceConfig
from azureml.core.runconfig import RunConfiguration
from azureml.core.webservice import AksWebservice, AciWebservice, Webservice
from azureml.pipeline.core.graph import PipelineParameter
from azureml.pipeline.steps import PythonScriptStep
from azureml.pipeline.core import Pipeline

import yaml
import io
import argparse
import json
import os
import mlflow


def IfRunInAzureML():
    try:
        run = Run.get_context(allow_offline=False)
        print('runrun')
        print(run)
    except:
        return False
    
    return run is not None

def RegisterModel(model_path, description, args):

    if IfRunInAzureML():
        experiment = Run.get_context(allow_offline=False).experiment
        ws = experiment.workspace

        Model.register(model_path = model_path,
                       model_name = args.operationId,
                       description = description,
                       workspace = ws,
                       tags={'userId': args.userId, 
                        'productName': args.productName, 
                        'deploymentName': args.deploymentName, 
                        'apiVersion':args.apiVersion,
                        'subscriptionId':args.subscriptionId})
    else:
        print('run on MLflow')
        mlFlowRun = mlflow.active_run()
        print(mlFlowRun)
        if mlFlowRun:
            mlflow.log_artifacts(model_path, model_path)
            model_uri = "runs:/{run_id}/{artifact_path}".format(run_id=mlFlowRun.info.run_id, artifact_path=model_path)
            print(model_uri)
            result = mlflow.register_model(
                model_uri,
                args.operationId
            )
            print(result)

    
def DownloadModel(model_id, model_path):
    if IfRunInAzureML():
        experiment = Run.get_context(allow_offline=False).experiment
        ws = experiment.workspace
        model = Model(ws, model_id)
        model.download(target_dir = model_path, exist_ok=True)

def DownloadOutputFromPredecessorOperation(predecessor_op_id, folder_path):
    return

def GetModelPath():
    # get model path if the model is deployed using AML
    if os.getenv('AZUREML_MODEL_DIR'):
        return os.getenv('AZUREML_MODEL_DIR')

def ParseArguments(run_type):
    parser = argparse.ArgumentParser(run_type)

    parser.add_argument("--runMode", type=str, help="run mode, support azureML or MLflow")
    parser.add_argument("--userInput", type=str, help="input data")
    parser.add_argument("--modelId", type=str, help="model key")

    if run_type == 'inference':
        parser.add_argument("--operationId", type=str, help="run id")
    else:
        parser.add_argument("--userId", type=str, help="user id")
        parser.add_argument("--productName", type=str, help="product name")
        parser.add_argument("--deploymentName", type=str, help="deployment name")
        parser.add_argument("--apiVersion", type=str, help="api version")
        parser.add_argument("--subscriptionId", type=str, help="subscription id")

        if run_type == 'deployment':
            parser.add_argument("--endpointId", type=str, help="endpoint id")

    args = parser.parse_args()
    userInput = json.loads(args.userInput)

    return args, userInput


def Init(luna_config_file='luna_config.yml'):
    with open(luna_config_file) as file:
        luna_config = yaml.full_load(file)
    return luna_config

def GetDeploymentConfig(luna_config, dns_name_label,tags):
    with open(luna_config['azureml']['workspace_config']) as file:
        documents = yaml.full_load(file)
        deployment_target = documents['deployment_target']
        aks_cluster = documents['aks_cluster']


    with open(luna_config['deploy_config']) as file:
        documents = yaml.full_load(file)

        if deployment_target == 'aci':
            deployment_config = AciWebservice.deploy_configuration()
            deployment_config.__dict__.update(documents['azureContainerInstance'])
            deployment_config.dns_name_label = dns_name_label
        elif deployment_target == 'aks':
            deployment_config = AksWebservice.deploy_configuration()
            deployment_config.__dict__.update(documents['kubernetes'])
            deployment_config.compute_target_name = aks_cluster
            deployment_config.namespace = dns_name_label

        deployment_config.tags = tags
    return deployment_config

def DeployModel():

    luna_config = Init()

    args, userInput = ParseArguments("deployment")

    dns_name_label = userInput["dns_name_label"]
    model_id = args.predecessorOperationId
    endpoint_id = args.operationId

    print(dns_name_label)

    ## If it is running in AML context
    if IfRunInAzureML():
        experiment = Run.get_context(allow_offline=False).experiment
        ws = experiment.workspace
        model = Model(ws, model_id)

        print(model)

        myenv = Environment.from_conda_specification('scoring', luna_config['conda_env'])

        print(luna_config['code']['score'])
        print(os.getcwd())

        inference_config = InferenceConfig(entry_script=luna_config['code']['score'], source_directory = os.getcwd(), environment=myenv)

        deployment_config = GetDeploymentConfig(
            luna_config, 
            dns_name_label,
            tags={'userId': args.userId, 
                'productName': args.productName, 
                'deploymentName': args.deploymentName, 
                'apiVersion':args.apiVersion,
                'subscriptionId':args.subscriptionId})
        
        service = Model.deploy(ws, endpoint_id, [model], inference_config, deployment_config)
        service.wait_for_deployment(show_output = True)

def RunProject(azureml_workspace, entry_point, experiment_name, parameters, tags):

    luna_config = Init()
    if azureml_workspace:
        run_config = RunConfiguration.load(luna_config['azureml']['run_config'])

        arguments = GetPipelineArguments(luna_config['MLproject'], entry_point, parameters)

        trainStep = PythonScriptStep(
            script_name=luna_config['code'][entry_point],
            arguments=arguments,
            inputs=[],
            outputs=[],
            source_directory=os.getcwd(),
            runconfig=run_config,
            allow_reuse=False
        )

        pipeline = Pipeline(workspace=azureml_workspace, steps=[trainStep])

        experiment = Experiment(azureml_workspace, experiment_name)
        pipeline_run = experiment.submit(pipeline, tags=tags)
        return pipeline_run.id
    
    else:
        return '00000000-0000-0000-0000-000000000000'

def GetPipelineArguments(mlproject_file_path, run_type, parameters):
    with open(mlproject_file_path) as file:
        documents = yaml.full_load(file)
        arguments = []
        for param in documents['entry_points'][run_type]['parameters']:
            argumentName = '--' + param
            arguments.append(argumentName)
            # get input parameter values, set default value if not provided
            if param in parameters:
                pipelineParam = PipelineParameter(name=param, default_value=parameters[param])
            else:
                pipelineParam = PipelineParameter(name=param, default_value=documents['entry_points'][run_type]['parameters'][param]['default'])
            arguments.append(pipelineParam)
        return arguments

def GetOperationNameByVerb(operationVerb):
    
    luna_config = Init()
    with open(luna_config['MLproject']) as file:
        documents = yaml.full_load(file)
        for operation in documents['entry_points']:
            if documents['entry_points'][operation]['verb'] == operationVerb:
                return operation

    return None


def GetOperationNameByNoun(operationNoun):
    
    luna_config = Init()
    with open(luna_config['MLproject']) as file:
        documents = yaml.full_load(file)
        for operation in documents['entry_points']:
            if documents['entry_points'][operation]['noun'] == operationVerb:
                return operation

    return None


from luna.logging.baseLunaLogger import BaseLunaLogger
from azureml.core import Workspace, Run, Experiment

class AzureMLLunaLogger(BaseLunaLogger):
    
    def log_metric(self, key, value):
        run = Run.get_context(allow_offline=False)
        run.log(key, value)

    def log_metrics(self, metrics):
        run = Run.get_context(allow_offline=False)
        for key in metrics:
            run.log(key, metrics[key])

    def upload_artifacts(self, local_path, upload_path):
        run = Run.get_context(allow_offline=False)
        run.upload_file(upload_path, local_path)


import yaml
import io
import argparse
import json
import os
import requests
import time


TRAINING_URL_FORMAT = "{base_url}/train?api-version={api_version}"
BATCHINFERENCE_URL_FORMAT = "{base_url}/models/{model_id}/batchinference?api-version={api_version}"
DEPLOY_URL_FORMAT = "{base_url}/models/{model_id}/deploy?api-version={api_version}"
GET_TRAINING_OP_URL_FORMAT = "{base_url}/operations/training/{model_id}?api-version={api_version}"
GET_INFERENCE_OP_URL_FORMAT = "{base_url}/operations/inference/{operation_id}?api-version={api_version}"
GET_DEPLOYMENT_OP_URL_FORMAT = "{base_url}/operations/deployment/{endpoint_id}?api-version={api_version}"

GET_MODEL_URL_FORMAT = "{base_url}/models/{model_id}?api-version={api_version}"
GET_ENDPOINT_URL_FORMAT = "{base_url}/endpoints/{endpoint_id}?api-version={api_version}"

class LunaClient(object):
    
    def __init__(self, base_url, key, api_version):
        self._base_url = base_url
        self._key = key
        self._api_version = api_version


    def get_request_header(self):
        return {"Accept": "application/json", "Ocp-Apim-Subscription-Key": self._key}

    def train_model(self, user_input):
        training_url = TRAINING_URL_FORMAT.format(base_url=self._base_url, api_version=self._api_version)
        response = requests.post(training_url, headers=self.get_request_header(), data=json.dumps(user_input))

        if response.status_code == 200:
            return response.json()['modelId']

        return None

    def get_training_operation(self, model_id):
        train_op_url = GET_TRAINING_OP_URL_FORMAT.format(base_url = self._base_url, model_id=model_id, api_version=self._api_version)
        response = requests.get(train_op_url, headers=self.get_request_header())
        if response.status_code == 200:
            return response.json()

        return None
        
    def get_batch_inference_operation(self, operation_id):
        inference_op_url = GET_INFERENCE_OP_URL_FORMAT.format(base_url = self._base_url, operation_id=operation_id, api_version=self._api_version)
        response = requests.get(inference_op_url, headers=self.get_request_header())
        if response.status_code == 200:
            return response.json()

        return None
    
    def get_deployment_operation(self, endpoint_id):
        inference_op_url = GET_DEPLOYMENT_OP_URL_FORMAT.format(base_url = self._base_url, endpoint_id=endpoint_id, api_version=self._api_version)
        response = requests.get(inference_op_url, headers=self.get_request_header())
        if response.status_code == 200:
            return response.json()

        return None

    def get_model(self, model_id):
        model_url = GET_MODEL_URL_FORMAT.format(base_url=self._base_url, model_id=model_id, api_version=self._api_version)
        response = requests.get(model_url, headers=self.get_request_header())
        if response.status_code == 200:
            return response.json()

    
    def get_deployed_endpoint(self, endpoint_id):
        model_url = GET_ENDPOINT_URL_FORMAT.format(base_url=self._base_url, endpoint_id=endpoint_id, api_version=self._api_version)
        response = requests.get(model_url, headers=self.get_request_header())
        if response.status_code == 200:
            return response.json()

    def wait_for_training_completion(self, model_id, timeout_is_seconds = 3600):
        while True:
            time.sleep(10)
            op = self.get_training_operation(model_id)
            if op is None:
                # TODO: what we gonna do here?
                continue
            if op["status"] != 'Running':
                if op["status"] == 'Completed':
                    return self.get_model(model_id)
                else:
                    return op
    
    def wait_for_batch_inference_completion(self, operation_id, timeout_is_seconds = 3600):
        while True:
            time.sleep(10)
            op = self.get_batch_inference_operation(operation_id)
            if op is None:
                # TODO: what we gonna do here?
                continue
            if op["status"] != 'Running':
                return op
    
    def wait_for_deployment_completion(self, endpoint_id, timeout_is_seconds = 3600):
        while True:
            time.sleep(10)
            op = self.get_deployment_operation(endpoint_id)
            if op is None:
                # TODO: what we gonna do here?
                continue
            if op["status"] != 'Running':
                return op
    
    def batch_inference(self, model_id, user_input):
        inference_url = BATCHINFERENCE_URL_FORMAT.format(base_url=self._base_url, model_id=model_id, api_version=self._api_version)
        response = requests.post(inference_url, headers=self.get_request_header(), data=json.dumps(user_input))
        
        if response.status_code == 200:
            return response.json()['operationId']

        return None

    def deploy_model(self, model_id, dns_name_label):
        deploy_url = DEPLOY_URL_FORMAT.format(base_url=self._base_url, model_id=model_id, api_version=self._api_version)
        user_input = {"dns_name_label": dns_name_label}
        response = requests.post(deploy_url, headers=self.get_request_header(), data=json.dumps(user_input))

        if response.status_code == 200:
            return response.json()['endpointId']

        return None

    def score_with_deployed_endpoint(self, endpoint_id, records, verboseMode=False):
        endpoint = self.get_deployed_endpoint(endpoint_id)
        header = {"Content-Type": "application/json", "Authorization": "Bearer "+endpoint["primaryKey"]}

        userInput = {"records": records, "verboseMode": verboseMode}

        response = requests.post(endpoint["scoringUrl"], headers=header, data=json.dumps(userInput))

        return response.text




from luna.logging.baseLunaLogger import BaseLunaLogger

import mlflow
import os

class MLFlowLunaLogger(BaseLunaLogger):
    
    def log_metric(self, key, value):
        mlflow.log_metric(key, value)

    def log_metrics(self, metrics):
        mlflow.log_metrics(metrics)

    def upload_artifacts(self, local_path, upload_path):
        if os.path.isdir(local_path):
            mlflow.log_artifacts(local_path, upload_path)
        elif os.path.isfile(local_path):
            mlflow.log_artifact(local_path, upload_path)


from azureml.core import Workspace, Run, Experiment
from azureml.core.compute import ComputeTarget, AmlCompute
from azureml.core.environment import Environment
from azureml.core.model import Model, InferenceConfig
from azureml.core.runconfig import RunConfiguration
from azureml.core.webservice import AksWebservice, AciWebservice, Webservice
from azureml.pipeline.core.graph import PipelineParameter
from azureml.pipeline.steps import PythonScriptStep
from azureml.pipeline.core import Pipeline

from luna.baseLunaUtils import BaseLunaUtils
from luna.logging.azuremlLunaLogger import AzureMLLunaLogger

import os
import yaml

class AzureMLLunaUtils(BaseLunaUtils):

    def Init(self, luna_config, run_mode, args, userInput):
        super().Init(luna_config, run_mode, args, userInput)
        self._logger = AzureMLLunaLogger()

    def GetAMLWorkspace(self):
        try:
            run = Run.get_context(allow_offline=False)
            return run.experiment.workspace
        except:
            return Workspace.from_config(path=self._luna_config["azureml"]["test_workspace_path"], 
                _file_name=self._luna_config["azureml"]["test_workspace_file_name"])
    
    def RegisterModel(self, model_path, description, luna_python_model=None):
        ws = self.GetAMLWorkspace()

        Model.register(model_path = model_path,
                       model_name = self._args.operationId,
                       description = description,
                       workspace = ws,
                       tags={'userId': self._args.userId, 
                        'productName': self._args.productName, 
                        'deploymentName': self._args.deploymentName, 
                        'apiVersion':self._args.apiVersion,
                        'subscriptionId':self._args.subscriptionId,
                        'modelId': self._args.operationId})

    def GetDeploymentConfig(self, tags, deployment_target=None, aks_cluster=None):

        # Read default deployment target and aks cluster info from the config files
        if not deployment_target:
            with open(self._luna_config['azureml']['workspace_config']) as file:
                documents = yaml.full_load(file)
                deployment_target = documents['deployment_target']
                if deployment_target == 'aks':
                    aks_cluster = documents['aks_cluster']

        with open(self._luna_config['deploy_config']) as file:
            documents = yaml.full_load(file)

            if deployment_target == 'aci':
                deployment_config = AciWebservice.deploy_configuration()
                deployment_config.__dict__.update(documents['azureContainerInstance'])
                deployment_config.dns_name_label = self._userInput["dns_name_label"]
            elif deployment_target == 'aks':
                deployment_config = AksWebservice.deploy_configuration()
                deployment_config.__dict__.update(documents['kubernetes'])
                deployment_config.compute_target_name = aks_cluster
                deployment_config.namespace = self._userInput["dns_name_label"]

            deployment_config.tags = tags
        return deployment_config

    def DeployModel(self):
        
        ws = self.GetAMLWorkspace()
        model = Model(ws, self._args.predecessorOperationId)
        myenv = Environment.from_conda_specification('scoring', self.luna_config['conda_env'])

        inference_config = InferenceConfig(entry_script=self.luna_config['code']['score'], source_directory = os.getcwd(), environment=myenv)

        deployment_config = self.GetDeploymentConfig(
            tags={'userId': self._args.userId, 
                'productName': self._args.productName, 
                'deploymentName': self._args.deploymentName, 
                'apiVersion':self._args.apiVersion,
                'subscriptionId':self._args.subscriptionId,
                'modelId': self._args.predecessorOperationId})
        
        service = Model.deploy(ws, self._args.endpointId, [model], inference_config, deployment_config)
        service.wait_for_deployment(show_output = True)

    def DownloadModel(self, model_path=""):
        ws = self.GetAMLWorkspace()
        model = Model(ws, self._args.predecessorOperationId)
        full_model_path = os.path.join(os.getcwd(), model_path, "models/artifacts")
        
        os.makedirs(full_model_path, exist_ok=True)
        model.download(target_dir = full_model_path, exist_ok=True)
        return full_model_path
